/*
Test of "ILSI" parallel linear algebra library.

This program solves the 1D heat equation via the conjugate gradient method.
Certainly not the most efficient solver (multigrid is way faster!), 
but general-purpose and easy to follow.

Dr. Orion Sky Lawlor, lawlor@alaska.edu, 2010-08-02 (Public Domain)
*/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <mpi.h>

#include "charm++.h" /* for CkPrintf */
#include <vector>

#include "ilsi.h" /* Charm++ iterative linear solver interface */
#include "ilsic.h"

#define trace(x) x // include debug printouts
//#define trace(x) /* empty; no tracing */

class myMatVec : public ILSI_Comm {
public:
	int n; /* number of local floats */
	int rank,size; /* MPI */
	myMatVec(int n_,int rank_,int size_) :n(n_),rank(rank_),size(size_) {}	

        /// Compute dest = A src, where A is the square matrix.
        ///  This product must take into account values from other
        ///  chunks, as well as this one.  This is a collective call.
        ///  It is not valid for src to equal dest.
        virtual void matrixVectorProduct(const double *src,double *dest)  { 
		static int itercount=0;
		itercount++;
		if (rank==0) {
			CkPrintf("--------- conjugate gradient iteration %d --------\n",itercount);
		}
		
		trace(CkPrintf("(vp %d) matrix-vector product begins\n",rank));
		// Request boundaries
		double leftghost=0.0;
		double rightghost=0.0;
		const int DOWNGOING_TAG=1234, UPGOING_TAG=4321;
		MPI_Status sts;
		if (rank-1>=0) { /* have a left neighbor */
			MPI_Send((void *)&src[0],1,MPI_DOUBLE, rank-1,DOWNGOING_TAG,MPI_COMM_WORLD);
			MPI_Recv(&leftghost,1,MPI_DOUBLE, rank-1,UPGOING_TAG,MPI_COMM_WORLD,&sts);
		}
		if (rank+1<size) { /* have a right neighbor */
			MPI_Send((void *)&src[n-1],1,MPI_DOUBLE, rank+1,UPGOING_TAG,MPI_COMM_WORLD);
			MPI_Recv(&rightghost,1,MPI_DOUBLE, rank+1,DOWNGOING_TAG,MPI_COMM_WORLD,&sts);
		}
		
		// Compute boundaries
		//  (note this would be simpler if we made a copy of src including the ghosts...)
		dest[0]=leftghost-2.0*src[0]+src[1];
		dest[n-1]=src[n-2]-2.0*src[n-1]+rightghost;
		
		// Compute interior
		for (int i=1;i<n-1;i++) {
			dest[i]=src[i-1]-2.0*src[i]+src[i+1];
		}
		trace(CkPrintf("(vp %d) matrix-vector product done\n",rank));
	}
       
        /// Do a global dot product of these two vectors. 
        /// This dot product must take into account values from other
        /// processors.  It is valid for a to equal b.
        /// All chunks are guaranteed to get the same return value.
        virtual double dotProduct(const double *a,const double *b) { 
		trace(CkPrintf("(vp %d) dot product begins\n",rank));
		double sum=0.0;
		for (int i=0;i<n;i++) sum+=a[i]*b[i];
		double gsum=0.0;
		MPI_Allreduce(&sum,&gsum,1,MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD); 
		trace(CkPrintf("(vp %d) dot product, returning %f\n",rank,gsum));
		return gsum;
	}
};

int main(int argc,char **argv)
{
	int rank,size;
	MPI_Init(&argc,&argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &size);

	int n=50; // size of the entire mesh
	if (argc>1) n=atoi(argv[1]);
	n=n/size; // number of vector elements that live on my processor
	if (n<1) n=1;

	ILSI_Param param; // contains tolerances, iteration count, etc
	ILSI_Param_new(&param);

	std::vector<double> rhs(n), lhs(n);
	for (int i=0;i<n;i++) {
		rhs[i]=0.0;
		lhs[i]=0.0;
	}
	/* apply external boundary conditions */
	if (rank==0) lhs[0]=100.0;
	if (rank==size-1) lhs[n-1]=-100.0;

	/* solve for the unknown right-hand side that produces this left-hand side via our matrix: 
	    lhs = matrix * rhs
	    lhs is known
	    rhs is unknown
	*/
	myMatVec mat(n,rank,size);
	ILSI_CG_Solver(&param,&mat,n,&lhs[0],&rhs[0]);


	MPI_Barrier(MPI_COMM_WORLD);
	if(rank == 0) {
		printf("IFEM test passed\n");
	}
	printf("Result at rank %d is %.2f .. %.2f\n",rank,rhs[0],rhs[n-1]);
	MPI_Finalize();
	return 0;
}
